/* -*- mode: c++; tab-width: 8; c-file-style: "gnu" -*- */
/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998, 1999, 2000, 2001  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@libsdl.org
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_QWin.cc,v 1.3 2002/09/01 00:37:24 slouken Exp $";
#endif
#include <stdio.h>
#include "SDL_bvddQWin.h"
#include <qapplication.h>
#include <qpe/qpeapplication.h>
#include <qdirectpainter_qws.h>
#include <qpe/config.h>
#include <qpe/qcopenvelope_qws.h>

extern int fb_hwrot;
extern int fb_direct;
static int mouse_button_mode=1;

screenRotationT screenRotation = SDL_QT_NO_ROTATION;

SDL_QWin::SDL_QWin(const QSize& size)
  : QWidget(0, "SDL_main"), my_painter(0),
    my_inhibit_resize(false), my_mouse_pos(-1,-1), my_flags(0),
    my_has_fullscreen(false), my_locked(0),
    rotation_(0), qteRotation_(0), keyRotation_(0), qteKeyRotation_(0),
    isSLA300InputFix_(0)
{
  for (int i = 0; i < 5; i++) curAxis_[i] = 0;
  setBackgroundMode(NoBackground);

  Config qpeconfig("qpe");
  if (qpeconfig.isValid()) {
    qpeconfig.setGroup("Screensaver");
    if (qpeconfig.hasKey("Interval"))
      {
	intervalSuspend = qpeconfig.readNumEntry("Interval", 0);
	intervalLightOff = qpeconfig.readNumEntry("Interval_LightOff", 0);
	intervalDim = qpeconfig.readNumEntry( "Interval_Dim", 0 );
      }
  }

  //  setWFlags(WStyle_Customize | WStyle_NoBorderEx);

  toggleSuspend(false);
  
}

SDL_QWin::~SDL_QWin() {
  // Nothing to do yet.
}

void SDL_QWin::setImage(QImage *image) {
}

bool SDL_QWin::event(QEvent *e)
{
  if( e->type() == QEvent::WindowDeactivate ){
    QKeyEvent *ke = new QKeyEvent(QEvent::KeyPress, Qt::Key_Escape,
				  0, 0);
    QueueKey(ke, 1);
    return true;
  }
  return QWidget::event(e);

}

void SDL_QWin::closeEvent(QCloseEvent *e) {
  SDL_PrivateQuit();
  e->ignore();
}

void SDL_QWin::setMousePos(const QPoint &pos) {
  QPoint realPos;

  if (0) {
	realPos = pos;
	my_mouse_pos = realPos;
  }else {
	  switch (rotation_) {
	  case 0:
	    realPos = pos;
	    break;
	  case 1:
	    realPos.setX(pos.y());
	    realPos.setY(fbSize_.height() - 1 - pos.x());
	    break;
	  case 2:
	    realPos.setX(fbSize_.width() - 1 - pos.x());
	    realPos.setY(fbSize_.height() - 1 - pos.y());
	    break;
	  case 3:
	    realPos.setX(fbSize_.height() - 1 - pos.y());
	    realPos.setY(pos.x());
	    break;
	  };
	  /*
	  fprintf(stderr, "setMousePos: realPos(%d, %d) rotation %d\n", realPos.x(), realPos.y(),rotation_);
	  fprintf(stderr, "setMousePos: fb(%d, %d) rotation %d\n", fbSize_.width(),fbSize_.height(),rotation_);
	  fprintf(stderr, "setMousePos: my(%d, %d) rotation %d\n", my_image.width(),my_image.height(),rotation_);
	  
	  
	  int borderWidth = (fbSize_.width() - my_image.width()) >> 1;
	  int borderHeight = (fbSize_.height() - my_image.height()) >> 1;
	  fprintf(stderr, "setMousePos: border(%d, %d) rotation %d\n", borderWidth,borderHeight,rotation_);
	  
	  if ( rotation_ & 1 ){
	    borderWidth = (fbSize_.height() - my_image.width()) >> 1;
	    borderHeight = (fbSize_.width() - my_image.height()) >> 1;
	  }
	  realPos.setX(realPos.x() - borderWidth);
	  realPos.setY(realPos.y() - borderHeight);
	  
	  fprintf(stderr, "setMousePos: realPos2(%d, %d)\n", realPos.x(), realPos.y());
	  */
	  switch (rotation_) {
	  case 0:
	    my_mouse_pos = realPos;
	    break;
	  case 1:
	    my_mouse_pos.setX(realPos.y());
	    my_mouse_pos.setY(my_image.height() - 1 - realPos.x());
	    break;
	  case 2:
	    my_mouse_pos.setX(my_image.width() - 1 - realPos.x());
	    my_mouse_pos.setY(my_image.height() - 1 - realPos.y());
	    break;
	  case 3:
	    my_mouse_pos.setX(realPos.y());
	    my_mouse_pos.setY( fbSize_.height() - realPos.x() - 10); //Need fix 
	    break;
	  }
  }

  //fprintf(stderr, "setMousePos: mymousePos(%d, %d)\n", (short)my_mouse_pos.x(), (short)my_mouse_pos.y());


}

void SDL_QWin::mouseMoveEvent(QMouseEvent *e) {
  Qt::ButtonState button = e->button();
  int sdlstate = 0;
  if( (button & Qt::LeftButton)) {
    sdlstate |= SDL_BUTTON_LMASK;
  }
  if( (button & Qt::RightButton)) {
    sdlstate |= SDL_BUTTON_RMASK;
  }
  if( (button & Qt::MidButton)) {
    sdlstate |= SDL_BUTTON_MMASK;
  }
   setMousePos(e->pos());
   SDL_PrivateMouseMotion(sdlstate, 0, my_mouse_pos.x(), my_mouse_pos.y());
   
}

void SDL_QWin::mousePressEvent(QMouseEvent *e) {
  mouseMoveEvent(e);
  Qt::ButtonState button = e->button();
  SDL_PrivateMouseButton(SDL_PRESSED,mouse_button_mode,
			 my_mouse_pos.x(), my_mouse_pos.y());
}

void SDL_QWin::mouseReleaseEvent(QMouseEvent *e) {
  setMousePos(e->pos());
  Qt::ButtonState button = e->button();
  SDL_PrivateMouseButton(SDL_RELEASED,mouse_button_mode,
			 my_mouse_pos.x(), my_mouse_pos.y());
  my_mouse_pos = QPoint(-1, -1);
  
}


int SDL_QWin::ApplyKeyRotation(int key)
{
  int c;
  int sdlScancode[] = { SDLK_LEFT, SDLK_DOWN, SDLK_RIGHT, SDLK_UP };

  switch (key) {
  case Qt::Key_Left:
    c = 0;
    break;
  case Qt::Key_Down:
    c = 1;
    break;
  case Qt::Key_Right:
    c = 2;
    break;
  case Qt::Key_Up:
    c = 3;
    break;
  default:
    return 0;
  }

//   c = (c + qteKeyRotation_) & 3;
//   return sdlScancode[(c - keyRotation_) & 3];
  return sdlScancode[(c + qteKeyRotation_ - keyRotation_) & 3];
}

/* Function to translate a keyboard transition and queue the key event
 * This should probably be a table although this method isn't exactly
 * slow.
 */
void SDL_QWin::QueueKey(QKeyEvent *e, int pressed)
{
  if (e->isAutoRepeat())
    return;

  SDL_keysym keysym;
  int scancode = e->key();

  /* Set the keysym information */
  if(scancode >= 'A' && scancode <= 'Z') {
    // Qt sends uppercase, SDL wants lowercase
    keysym.sym = static_cast<SDLKey>(scancode + 32);
  } else if(scancode  >= 0x1000) {
    // Special keys
    switch(scancode) {
    case Qt::Key_Escape: scancode = SDLK_ESCAPE; break;
    case Qt::Key_Tab: scancode = SDLK_TAB; break;
    case Qt::Key_Backspace: scancode = SDLK_BACKSPACE; break;
    case Qt::Key_Return: scancode = SDLK_RETURN; break;
    case Qt::Key_Enter: scancode = SDLK_KP_ENTER; break;
    case Qt::Key_Insert: scancode = SDLK_INSERT; break;
    case Qt::Key_Delete: scancode = SDLK_DELETE; break;
    case Qt::Key_Pause: scancode = SDLK_PAUSE; break;
    case Qt::Key_Print: scancode = SDLK_PRINT; break;
    case Qt::Key_SysReq: scancode = SDLK_SYSREQ; break;
    case Qt::Key_Home: scancode = SDLK_HOME; break;
    case Qt::Key_End: scancode = SDLK_END; break;
    // We want the control keys to rotate with the screen
    case Qt::Key_Left:
    case Qt::Key_Up:
    case Qt::Key_Right:
    case Qt::Key_Down:
      scancode = ApplyKeyRotation(scancode);
      break;
    case Qt::Key_Prior: scancode = SDLK_PAGEUP; break;
    case Qt::Key_Next: scancode = SDLK_PAGEDOWN; break;
    case Qt::Key_Shift: scancode = SDLK_LSHIFT; break;
    case Qt::Key_Control: scancode = SDLK_LCTRL; break;
    case Qt::Key_Meta: scancode = SDLK_LMETA; break;
    case Qt::Key_Alt: scancode = SDLK_LALT; break;
    case Qt::Key_CapsLock: scancode = SDLK_CAPSLOCK; break;
    case Qt::Key_NumLock: scancode = SDLK_NUMLOCK; break;
    case Qt::Key_ScrollLock: scancode = SDLK_SCROLLOCK; break;
    case Qt::Key_F1: scancode = SDLK_F1; break;
    case Qt::Key_F2: scancode = SDLK_F2; break;
    case Qt::Key_F3: scancode = SDLK_F3; break;
    case Qt::Key_F4: scancode = SDLK_F4; break;
    case Qt::Key_F5: scancode = SDLK_F5; break;
    case Qt::Key_F6: scancode = SDLK_F6; break;
    case Qt::Key_F7: scancode = SDLK_F7; break;
    case Qt::Key_F8: scancode = SDLK_F8; break;
    case Qt::Key_F9: scancode = SDLK_F9; break;
    case Qt::Key_F10: 
      scancode = SDLK_F10;
      mouse_button_mode=1;
      break;
    case Qt::Key_F11:
      scancode = SDLK_F11;
      mouse_button_mode=3;
      break;
    case Qt::Key_F12:
      scancode = SDLK_F12;
      mouse_button_mode=2;
      break;
    case Qt::Key_F13: scancode = SDLK_F13; break;
    case Qt::Key_F14: scancode = SDLK_F14; break;
    case Qt::Key_F15: scancode = SDLK_F15; break;
    case Qt::Key_Super_L: scancode = SDLK_LSUPER; break;
    case Qt::Key_Super_R: scancode = SDLK_RSUPER; break;
    case Qt::Key_Menu: scancode = SDLK_MENU; break;
    case Qt::Key_Help: scancode = SDLK_HELP; break;

    case 0x5001:                                //zaurus zen key Qt::Key_F21
    case 0x5002: scancode = SDLK_LSUPER; break; //zaurus zen key Qt::Key_F25
    case 0x5003:                                //zaurus kana key Qt::Key_F26
    case 0x5004: scancode = SDLK_RSUPER; break; //zaurus kana key Qt::Key_F31
    case Qt::Key_F22: scancode = SDLK_MENU; break;   //zaurus Fn key

    case Qt::Key_F33:
      // FIXME: This is a hack to enable the OK key on
      // Zaurii devices. SDLK_RETURN is a suitable key to use
      // since it often is used as such.
      //     david@hedbor.org
      scancode = SDLK_RETURN;
      break;

    default:
      scancode = SDLK_UNKNOWN;
      break;
    }
    keysym.sym = static_cast<SDLKey>(scancode);
  } else {
    keysym.sym = static_cast<SDLKey>(scancode);
  }
  keysym.scancode = scancode;
  keysym.mod = KMOD_NONE;
  ButtonState st = e->state();
  if( (st & ShiftButton) )   { keysym.mod = static_cast<SDLMod>(keysym.mod | KMOD_LSHIFT);  }
  if( (st & ControlButton) ) { keysym.mod = static_cast<SDLMod>(keysym.mod | KMOD_LCTRL);  }
  if( (st & AltButton) )     { keysym.mod = static_cast<SDLMod>(keysym.mod | KMOD_LALT);  }
  if ( SDL_TranslateUNICODE ) {
    QChar qchar = e->text()[0];
    keysym.unicode = qchar.unicode();
  } else {
    keysym.unicode = 0;
  }

  /* NUMLOCK and CAPSLOCK are implemented as double-presses in reality */
  //	if ( (keysym.sym == SDLK_NUMLOCK) || (keysym.sym == SDLK_CAPSLOCK) ) {
  //		pressed = 1;
  //	}

  if (isSLA300InputFix_ &&
      (keysym.sym == SDLK_SPACE || (keysym.sym >= 273 && keysym.sym <= 276))) {
      if (keysym.sym >= 273) curAxis_[keysym.sym-273] = pressed;
      else curAxis_[4] = pressed;
  }

  /* Queue the key event */
  if ( pressed ) {
      /* fprintf(stderr, "press %d\n", keysym.sym); */
      SDL_PrivateKeyboard(SDL_PRESSED, &keysym);
      if (isSLA300InputFix_ &&
          (keysym.sym >= 273 && keysym.sym <= 276)) {
          if (keysym.sym == SDLK_UP) {
              if (curAxis_[1]) {
                  keysym.sym = SDLK_DOWN;
                  keysym.scancode = Qt::Key_Down;
                  curAxis_[1] = 0;
                  /* fprintf(stderr, "force release %d\n", keysym.sym); */
                  SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
              }
          }
          else if (keysym.sym == SDLK_DOWN) {
              if (curAxis_[0]) {
                  keysym.sym = SDLK_UP;
                  keysym.scancode = Qt::Key_Up;
                  curAxis_[0] = 0;
                  /* fprintf(stderr, "force release %d\n", keysym.sym); */
                  SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
              }
          }
          else if (keysym.sym == SDLK_RIGHT) {
              if (curAxis_[3]) {
                  keysym.sym = SDLK_LEFT;
                  keysym.scancode = Qt::Key_Left;
                  curAxis_[3] = 0;
                  /* fprintf(stderr, "force release %d\n", keysym.sym); */
                  SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
              }
          }
          else if (keysym.sym == SDLK_LEFT) {
              if (curAxis_[2]) {
                  keysym.sym = SDLK_RIGHT;
                  keysym.scancode = Qt::Key_Right;
                  curAxis_[2] = 0;
                  /* fprintf(stderr, "force release %d\n", keysym.sym); */
                  SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
              }
          }
      }
  } else {
      /* fprintf(stderr, "release %d\n", keysym.sym); */
      SDL_PrivateKeyboard(SDL_RELEASED, &keysym);

      if (isSLA300InputFix_ &&
          (keysym.sym == SDLK_SPACE ||
           (keysym.sym >= 273 && keysym.sym <= 276))) {
          for (int i = 0; i < 4; i++) {
              if (curAxis_[i]) {
                  int sym = i+273;
                  keysym.sym = static_cast<SDLKey>(sym);
                  if (sym == SDLK_UP) keysym.scancode = Qt::Key_Up;
                  else if (sym == SDLK_RIGHT) keysym.scancode = Qt::Key_Right;
                  else if (sym == SDLK_DOWN) keysym.scancode = Qt::Key_Down;
                  else if (sym == SDLK_LEFT) keysym.scancode = Qt::Key_Left;
                  curAxis_[i] = 0;
                  /* fprintf(stderr, "force release %d\n", keysym.sym); */
                  SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
              }
          }
          if (curAxis_[4]) {
              keysym.scancode = keysym.sym = SDLK_SPACE;
              curAxis_[4] = 0;
              /* fprintf(stderr, "force release %d\n", keysym.sym); */
              SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
          }
      }
  }
}

void SDL_QWin::toggleSuspend(bool enable)
{
    QCopEnvelope e("QPE/System", "setScreenSaverIntervals(int,int,int)" );
    if (enable)
    {
      printf("Enable Suspend\n");
      e << intervalDim << intervalLightOff << intervalSuspend;
    }else {
      printf("Disable Suspend\n");
      e << 0 << 0 << 0;
    }
}



