/* -*- mode: c++; tab-width: 8; c-file-style: "gnu" -*- */
/*
  SDL - Simple DirectMedia Layer
  Copyright (C) 1997, 1998, 1999, 2000, 2001, 2002  Sam Lantinga

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Sam Lantinga
  slouken@libsdl.org
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_sysvideo.cc,v 1.3 2002/09/01 00:37:24 slouken Exp $";
#endif

/* Qtopia based framebuffer implementation */

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
extern int errno;

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <linux/fb.h>
#include <sys/mman.h>
#include <asm/page.h>	
#include <stdint.h>	
 
#include <qapplication.h>
#include <qpe/qpeapplication.h>
#include <qpe/qcopenvelope_qws.h>
#include <qgfx_qws.h>
#include <qwindowsystem_qws.h>
#include <qwidget.h>
#include <qwidgetlist.h>
#include <qdirectpainter_qws.h>
#include <qwsdisplay_qws.h>

#include "SDL.h"
#include "SDL_timer.h"

#include "SDL_bvddQWin.h"
#include "SDL_sysvideo.h"

extern "C" {

#include "SDL_sysvideo.h"
#include "SDL_bvddsysmouse_c.h"
#include "SDL_bvddsysevents_c.h"
#include "SDL_events_c.h"
#include "SDL_bvddsyswm_c.h"
#include "SDL_bvddlowvideo.h"
#include "bvdd.h"

  // #define QTOPIA_DEBUG
#define QTOPIA_LOG
#define QT_HIDDEN_SIZE	32	/* starting hidden window size */

#ifdef QTOPIA_LOG
  static FILE *st_logfp = NULL;
#endif
  static inline void LOG(char *fmt, ...)
  {
#ifdef QTOPIA_LOG
    va_list ap;

    va_start(ap, fmt);
    vfprintf(st_logfp, fmt, ap);
    if (st_logfp != stderr) {
      fflush(st_logfp);
      vfprintf(stderr, fmt, ap);
    }
    va_end(ap);
#endif
  }


/* Name of the environment variable used to invert the screen rotation or not:
   Possible values:
   !=0 : Screen is 270- rotated
   0: Screen is 90- rotated*/
#define SDL_QT_ROTATION_ENV_NAME "SDL_QT_ROTATION"
#define SDL_QT_FORCE_QVGA "SDL_FORCE_QVGA"
//#define SDL_QT_NO_DOUBLE "SDL_NO_DOUBLE"

  /* Initialization/Query functions */
  static int QT_VideoInit(_THIS, SDL_PixelFormat *vformat);
  static SDL_Rect **QT_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags);
  static SDL_Surface *QT_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags);
  static void QT_UpdateMouse(_THIS);
  static int QT_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors);
  static void QT_VideoQuit(_THIS);

  /* Hardware surface functions */
  static int QT_AllocHWSurface(_THIS, SDL_Surface *surface);
  static int QT_LockHWSurface(_THIS, SDL_Surface *surface);
  static void QT_UnlockHWSurface(_THIS, SDL_Surface *surface);
  static void QT_FreeHWSurface(_THIS, SDL_Surface *surface);
  static int QT_FlipHWSurface(_THIS, SDL_Surface *surface);

  static int QT_ToggleFullScreen(_THIS, int fullscreen);

  static int QT_IconifyWindow(_THIS);
  static SDL_GrabMode QT_GrabInput(_THIS, SDL_GrabMode mode);

  static int bvdd_fd;

#define LCD_SHARP_QVGA 0
#define LCD_SHARP_VGA  1

  static int st_lcd_mode = LCD_SHARP_VGA;
  static void change_resolution(int mode);

  static short *gfx_buffer;
  static int gfx_width;
  static int gfx_height;

  static short *bvdd_current_mem;
  static int bvdd_width;
  static int bvdd_height;

  static bvdd_vram_plane_t base_frame[3];
  static bvdd_frame_t st_bvdd_frame;
  static int flip_page;
  static void *bvdd_mapped_mem[3];

  static int sdlRotation=0;

  static int isKHloaded;

  static int flip_num = 0;
  static int updateRect_num = 0;
  static int no_flip = 0;
  static int no_updateRects = 0;

#define SDL_DIRECT	0x04000000	/* Direct Update flag */
 
  /* FB driver bootstrap functions */

  static int QT_Available(void)
  {
    return(1);
  }

  static void QT_DeleteDevice(SDL_VideoDevice *device)
  {
    free(device->hidden);
    free(device);
  }

  static SDL_VideoDevice *QT_CreateDevice(int devindex)
  {
    SDL_VideoDevice *device;

    /* Initialize all variables that we clean on shutdown */
    device = (SDL_VideoDevice *)malloc(sizeof(SDL_VideoDevice));
    if ( device ) {
      memset(device, 0, (sizeof *device));
      device->hidden = (struct SDL_PrivateVideoData *)
	malloc((sizeof *device->hidden));
    }
    if ( (device == NULL) || (device->hidden == NULL) ) {
      SDL_OutOfMemory();
      if ( device ) {
	free(device);
      }
      return(0);
    }
    memset(device->hidden, 0, (sizeof *device->hidden));

    /* Set the function pointers */
    device->VideoInit = QT_VideoInit;
    device->ListModes = QT_ListModes;
    device->SetVideoMode = QT_SetVideoMode;
    device->UpdateMouse = QT_UpdateMouse;
    device->SetColors = QT_SetColors;
    device->UpdateRects = NULL;
    device->VideoQuit = QT_VideoQuit;
    device->AllocHWSurface = QT_AllocHWSurface;
    device->CheckHWBlit = NULL;
    device->FillHWRect = NULL;
    device->SetHWColorKey = NULL;
    device->SetHWAlpha = NULL;
    device->LockHWSurface = QT_LockHWSurface;
    device->UnlockHWSurface = QT_UnlockHWSurface;
    device->FlipHWSurface = QT_FlipHWSurface;
    device->FreeHWSurface = QT_FreeHWSurface;
    device->SetIcon = NULL;
    device->SetCaption = QT_SetWMCaption;
    device->IconifyWindow = QT_IconifyWindow;
    device->GrabInput = QT_GrabInput;
    device->GetWMInfo = NULL;
    device->FreeWMCursor = QT_FreeWMCursor;
    device->CreateWMCursor = QT_CreateWMCursor;
    device->ShowWMCursor = QT_ShowWMCursor;
    device->WarpWMCursor = QT_WarpWMCursor;
    device->InitOSKeymap = QT_InitOSKeymap;
    device->PumpEvents = QT_PumpEvents;

    device->free = QT_DeleteDevice;
    device->ToggleFullScreen = QT_ToggleFullScreen;

    /* Set the driver flags */
    device->handles_any_size = 0;
	
    return device;
  }

  VideoBootStrap Qtopia_bootstrap = {
    "bvdd", "BVDD overlay graphics",
    QT_Available, QT_CreateDevice
  };
  
  static int alloc_plane(bvdd_vram_plane_t *vp, int type, int size, uint8_t c)
  {
    vp->area_type = type;
    vp->size = size;
    if (ioctl(bvdd_fd, BVDD_IOCTL_ALLOC_PLANE, vp) < 0) {
	printf("vo_bvdd: mmap failed.\n");
	return -1;
    }
    vp->addr.virt = mmap(0, vp->size, PROT_READ|PROT_WRITE, MAP_SHARED, bvdd_fd, vp->addr.phys);
    if (vp->addr.virt == MAP_FAILED) {
	printf("vo_bvdd: mmap failed.\n");
	return -1;
    }
    memset(vp->addr.virt, c, vp->size);
    return 1;
  }

  /* Function to sort the display_list */
  static int CompareModes(const void *A, const void *B)
  {
    return 0;
  }

  /* Yes, this isn't the fastest it could be, but it works nicely */
  static int QT_AddMode(_THIS, int index, unsigned int w, unsigned int h)
  {
    SDL_Rect *mode;
    int i;
    int next_mode;

    /* Check to see if we already have this mode */
    if ( SDL_nummodes[index] > 0 ) {
      for ( i=SDL_nummodes[index]-1; i >= 0; --i ) {
	mode = SDL_modelist[index][i];
	if ( (mode->w == w) && (mode->h == h) ) {
	  return(0);
	}
      }
    }

    /* Set up the new video mode rectangle */
    mode = (SDL_Rect *)malloc(sizeof *mode);
    if ( mode == NULL ) {
      SDL_OutOfMemory();
      return(-1);
    }
    mode->x = 0;
    mode->y = 0;
    mode->w = w;
    mode->h = h;
#ifdef QTOPIA_DEBUG
    fprintf(stderr, "Adding mode %dx%d at %d bytes per pixel\n", w, h, index+1);
#endif

    /* Allocate the new list of modes, and fill in the new mode */
    next_mode = SDL_nummodes[index];
    SDL_modelist[index] = (SDL_Rect **)
      realloc(SDL_modelist[index], (1+next_mode+1)*sizeof(SDL_Rect *));
    if ( SDL_modelist[index] == NULL ) {
      SDL_OutOfMemory();
      SDL_nummodes[index] = 0;
      free(mode);
      return(-1);
    }
    SDL_modelist[index][next_mode] = mode;
    SDL_modelist[index][next_mode+1] = NULL;
    SDL_nummodes[index]++;

    return(0);
  }

  int QT_VideoInit(_THIS, SDL_PixelFormat *vformat)
  {
    const char *SDL_bvdddev;

     /* Initialize the QPE Application  */
     /* Determine the screen depth */
    vformat->BitsPerPixel = QPixmap::defaultDepth();

    // For now we hardcode the current depth because anything else
    // might as well be emulated by SDL rather than by Qtopia.

    //frame buffer device open.
    SDL_bvdddev = getenv("SDL_BVDDDEV");
    if ( SDL_bvdddev == NULL ) {
    	SDL_bvdddev = "/dev/bvdd";
    }
    bvdd_fd = open(SDL_bvdddev, O_RDWR, 0);
    if ( bvdd_fd < 0 ) {
    	SDL_SetError("Unable to open %s", SDL_bvdddev);
    	return(-1);
    }
    close(bvdd_fd);

    QSize desktop_size = qApp->desktop()->size();
    QT_AddMode(_this, ((vformat->BitsPerPixel+7)/8)-1,
	       desktop_size.width(), desktop_size.height());
    QT_AddMode(_this, ((vformat->BitsPerPixel+7)/8)-1,
	       desktop_size.height(), desktop_size.width());

    /* Create the window / widget */
    SDL_Win = new SDL_QWin(QSize(QT_HIDDEN_SIZE, QT_HIDDEN_SIZE));
    ((QPEApplication*)qApp)->showMainWidget(SDL_Win);
    /* Fill in some window manager capabilities */
    _this->info.wm_available = 0;

#ifdef QTOPIA_LOG
    st_logfp = fopen("/tmp/sdl-qt-debug", "w");
    if (!st_logfp)
      st_logfp = stderr;
#endif

    QT_GrabInput(_this, SDL_GRAB_ON);

    // keyhelper loading for SDL

    isKHloaded=(fopen("/home/zaurus/Settings/keyhelper_SDL.xml","r") != NULL) ? 1 : 0;
    if (isKHloaded) {
	system("qcop QPE/KeyHelper \"reload(QString)\" keyhelper_SDL.xml"); //QcopEnvelope can't use. 
	LOG("keyhelper_SDL.xml loaded\n");
    }
    return(0);
  }

  /* We support any dimension at our bit-depth */
  SDL_Rect **QT_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
  {
    SDL_Rect **modes;

    modes = ((SDL_Rect **)0);
    if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
      modes = SDL_modelist[((format->BitsPerPixel+7)/8)-1];
    } else {
      if ( format->BitsPerPixel ==
	   _this->screen->format->BitsPerPixel ) {
	modes = ((SDL_Rect **)-1);
      }
    }
    return(modes);
  }

  /* Various screen update functions available */
  static void QT_NormalUpdate0(_THIS, int numrects, SDL_Rect *rects);
  static void QT_NormalUpdate3(_THIS, int numrects, SDL_Rect *rects);
  static void QT_DirectUpdate(_THIS, int numrects, SDL_Rect *rects);


  static int QT_SetFullScreen(_THIS, SDL_Surface *screen, int fullscreen)
  {
    return -1;
  }

  static int QT_ToggleFullScreen(_THIS, int fullscreen)
  {
    return -1;
  }

  static void QT_GetQteServerSpec(_THIS, int *rotation, bool *isQvga)
  {
    const char *user;
    char buf[FILENAME_MAX];
    FILE *fp;
    int rot = 0, is_qvga = 0;

    user = getenv("USER");
    snprintf(buf, sizeof(buf), "/tmp/qtembedded-%s/QtEmbedded-0.spec",
             user ? user : "root");
    LOG("QT_GetRotation: Read spec from '%s'\n", buf);

    fp = fopen(buf, "rb");
    if (fp) {
      int size;
      if ((size = fread(buf, 1, sizeof(buf) - 1, fp)) > 0) {
        buf[size] = '\0';
        LOG(" spec is '%s'\n", buf);
        // get rotation value
        if (strstr(buf, "Rot")) {
          rot = atoi(strstr(buf, "Rot") + 3);
          rot /= 90;
          if (rot < 0 || rot > 3) {
            rot = 0;
          }
        } else
          rot = 0;

        // get qvga mode in SL-C700
        if (strstr(buf, "Qvga"))
          is_qvga = 1;
      }
      fclose(fp);
    }

    LOG(" Rot=%d, Qvga=%d\n", rot, is_qvga);

    if (rotation)
      *rotation = rot;
    if (isQvga)
      *isQvga = is_qvga;
  }



  /* FIXME: check return values and cleanup here */
  SDL_Surface *QT_SetVideoMode(_THIS, SDL_Surface *current,
			       int width, int height, int bpp, Uint32 flags)
  {
    QImage *qimage;
    QSize qteSize = qApp->desktop()->size();
    QSize fbSize;
    QSize my_image;
    QSize userSize;
    int tmp_ioctl_data;

    int vram_size = 0;

    int qteRotation, userRotation;
    int qteKeyRotation, sdlKeyRotation;
    bool isQteQvga;
    bool force_qvga;

    // qte Ǥβž٤
    QT_GetQteServerSpec(_this, &qteRotation, &isQteQvga);

    // SL-C700 Υ
    bool isInputStyle = false;
    int status = system("/home/QtPalmtop/bin/chkhinge");
    if (WEXITSTATUS(status) != 2)
      isInputStyle = true;
    LOG("QT_SetVideoMode: SL-C700 Style is %s\n",
	isInputStyle ? "Input style" : "View style");

    if(isInputStyle)
      qteRotation = 3;

    // get force qvga env
    const char *envQVGA = getenv(SDL_QT_FORCE_QVGA);
    if(envQVGA){
      force_qvga = atoi(envQVGA) == 1 ? TRUE : FALSE;
    }else{
      force_qvga = FALSE;
    }

/*    get no double env
    const char *envNoDbl = getenv(SDL_QT_NO_DOUBLE);
    if(envNoDbl){
      int noDouble = atoi(envNoDbl);
      if(noDouble == 1)
	flags &= ~SDL_DOUBLEBUF;
    }
*/
    // get rotation env
    const char *envRotStr = getenv(SDL_QT_ROTATION_ENV_NAME);
    if(envRotStr){
      userRotation = atoi(envRotStr);
      sdlRotation = userRotation >= -1 ? userRotation : qteRotation;
    }else {
      if(flags & SDL_DIRECT){ // Screen direct draw flag.(SDL_DIRECT 0x04000000)
	userRotation = -1;
	sdlRotation = -1;
      }else{
	userRotation = 0;
	sdlRotation = qteRotation;
      }
    }

    LOG("QT_SetVideoMode: argSize=%dx%d\n", width, height);
    LOG("QT_SetVideoMode: qteSize=%dx%d\n", qteSize.width(), qteSize.height());
    LOG("QT_SetVideoMode: qteRotation=%d\n", qteRotation);
    LOG("QT_SetVideoMode: userRotation=%d\n", userRotation);
    LOG("QT_SetVideoMode: sdlRotation=%d\n", sdlRotation);
    
    
    
    // keyboard rotation
    qteKeyRotation = qteRotation;
    if (isQteQvga)
      qteKeyRotation = 3;
    
    sdlKeyRotation = qteKeyRotation;
    SDL_Win->setKeyRotation(sdlKeyRotation, qteKeyRotation);
    
    LOG("QT_SetVideoMode: qteKeyRotation=%d\n", qteKeyRotation);
    LOG("QT_SetVideoMode: sdlKeyRotation=%d\n", sdlKeyRotation);
    
    if ( flags & SDL_OPENGL ) {
      SDL_SetError("QT_SetVideoMode: OpenGL not supported");
      return(NULL);
    }

    // change resolution
    if( ((width <=240) && ( height <=320) && ( sdlRotation == 0 || sdlRotation == -1) ) || 
	((width <=320) && ( height <=240) && ( sdlRotation == 3 ) ) || 
	force_qvga){
      if(sdlRotation ==3){
	fbSize.setWidth(320);
	fbSize.setHeight(240);
      }else{
	fbSize.setWidth(240);
	fbSize.setHeight(320);
      }
      LOG("QT_SetVideoMode: Change LCD to QVGA mode.\n");
      change_resolution(LCD_SHARP_QVGA);
      vram_size = 240 * 320 * 2;
    }else{ 
      if(sdlRotation ==3){
	fbSize.setWidth(640);
	fbSize.setHeight(480);
      }else{
	fbSize.setWidth(480);
	fbSize.setHeight(640);
      }
      vram_size = 480 * 640 * 2;
    }

    // frame size setting
    if(sdlRotation == -1){
      current->w = fbSize.width();
      current->h = fbSize.height();
    }else{
      current->w = width;
      current->h = height;
    }


    // open bvdd device after change resolution.
    bvdd_fd = open("/dev/bvdd", O_RDWR);
    if (bvdd_fd < 0) {
	perror("open");
	exit(1) ;
    }
    LOG("QT_SetVideoMode: Open bvdd device\n");

    SDL_Win->setScreenRotation(sdlRotation, qteRotation);
    SDL_Win->setFBSize(fbSize);

    my_image.setWidth(width);
    my_image.setHeight(height);
    SDL_Win->setMyImage(my_image);

    if (alloc_plane(&base_frame[0], BVDD_VRAM_AREA_BVDD, vram_size, 0) < 0)
      exit(1);
    if(flags &SDL_DOUBLEBUF){
      if (alloc_plane(&base_frame[1], BVDD_VRAM_AREA_BVDD, vram_size, 0) < 0)
	exit(1);
      if (alloc_plane(&base_frame[2], BVDD_VRAM_AREA_BVDD, vram_size, 0) < 0)
	exit(1);
      LOG("QT_SetVideoMode: Using Double Buffering\n");
      flip_page = 0;
    }

    
    // base frame setting
    st_bvdd_frame.format = BVDD_FORMAT_RGB16;
    if(st_lcd_mode == LCD_SHARP_QVGA){
      st_bvdd_frame.width = 240;
      st_bvdd_frame.height = 320;
    }else{
      st_bvdd_frame.width = 480;
      st_bvdd_frame.height = 640;
    }
    st_bvdd_frame.x = 0;
    st_bvdd_frame.y = 0;
    st_bvdd_frame.plane_id = base_frame[flip_page].id;

    ioctl(bvdd_fd, BVDD_IOCTL_SETUP_BASE_FRAME, &st_bvdd_frame);

    bvdd_width = st_bvdd_frame.width;
    bvdd_height = st_bvdd_frame.height;
    
    bvdd_mapped_mem[0] = (void*)base_frame[0].addr.virt;
    bvdd_mapped_mem[1] = (void*)base_frame[1].addr.virt;
    bvdd_mapped_mem[2] = (void*)base_frame[2].addr.virt;
    
    LOG("QT_SetVideoMode: BVDD maped size %d address %x.\n", vram_size, bvdd_mapped_mem[0]);
    if(flags &SDL_DOUBLEBUF){
      LOG("QT_SetVideoMode: BVDD maped size %d address %x.\n", vram_size, bvdd_mapped_mem[1]);
      LOG("QT_SetVideoMode: BVDD maped size %d address %x.\n", vram_size, bvdd_mapped_mem[2]);
    }

    if(sdlRotation == -1){
      current->pitch = current->w * 2;
    }else{
      current->pitch = current->w * 2;
      gfx_buffer = (short *)malloc(width * height * 2);
      gfx_width = width;
      gfx_height = height;
    }

    current->flags = (SDL_FULLSCREEN|SDL_HWSURFACE |(flags&SDL_DOUBLEBUF));
    
    if(sdlRotation == -1 ) {
      current->pixels = (void *)(bvdd_mapped_mem[flip_page]);
      _this->UpdateRects = QT_DirectUpdate;
      LOG("QT_SetVideoMode: Using DirectPaint mode\n");
    } else if(sdlRotation == 0){
      current->pixels = (void *)gfx_buffer;
      bvdd_current_mem = (short *)(bvdd_mapped_mem[flip_page]);
      _this->UpdateRects = QT_NormalUpdate0;
      LOG("QT_SetVideoMode: Using Rotation 0\n");
    } else { // if(sdlRotation == 3){
      current->pixels =  (void *)gfx_buffer;
      bvdd_current_mem = (short *)(bvdd_mapped_mem[flip_page]);
      _this->UpdateRects = QT_NormalUpdate3;
      LOG("QT_SetVideoMode: Using Rotation 270\n");
    }
    
    //     fprintf(stderr,"QT_SetVideoMode() qImage:%dx%d %d\n",
    // 			qimage->width(), qimage->height(), qimage->bytesPerLine());
    /* We're done */
    return(current);
  }


  /* Update the current mouse state and position */
  void QT_UpdateMouse(_THIS)
  {
    QPoint point(-1, -1);
    if ( SDL_Win->isActiveWindow() ) {
      point = SDL_Win->mousePos();
    }
    
    if ( (point.x() >= 0) && (point.x() < SDL_VideoSurface->w) &&
	 (point.y() >= 0) && (point.y() < SDL_VideoSurface->h) ) {
      SDL_PrivateAppActive(1, SDL_APPMOUSEFOCUS);
      SDL_PrivateMouseMotion(0, 0,
			     (Sint16)point.x(), (Sint16)point.y());
    } else {
      SDL_PrivateAppActive(0, SDL_APPMOUSEFOCUS);
    }
  }

  /* We don't actually allow hardware surfaces other than the main one */
  static int QT_AllocHWSurface(_THIS, SDL_Surface *surface)
  {
    return(-1);
  }
  static void QT_FreeHWSurface(_THIS, SDL_Surface *surface)
  {
    return;
  }
  static int QT_LockHWSurface(_THIS, SDL_Surface *surface)
  {
    return(0);
  }
  static void QT_UnlockHWSurface(_THIS, SDL_Surface *surface)
  {
    return;
  }

  // rotation 0 degree
  static void QT_NormalUpdate0(_THIS, int numrects, SDL_Rect *rects)
  {
    updateRect_num++;
    if(updateRect_num >15 && flip_num < 5){    // App has no SDL_FLIP

      if( (_this->screen->flags & SDL_DOUBLEBUF) && !no_flip){
	st_bvdd_frame.plane_id = base_frame[0].id;
	ioctl(bvdd_fd, BVDD_IOCTL_BRANCH_BASE_FRAME, &st_bvdd_frame);
	/*
	if ( ioctl(bvdd_fd, BVDD_IOCTL_BRANCH_BASE_FRAME, &st_bvdd_frame)< 0 ) {
	  SDL_SetError("QT_UpdateRects: BVDD ioctl failed");
	  return;
	}
	*/
	bvdd_current_mem = (short*)bvdd_mapped_mem[0];
	printf("QT_UpdateRects: Not use SDL_Flip.\n");
	no_flip = 1;
      }
    }

    //printf("Rect %d flip %d\n",updateRect_num,flip_num);

    for(int i=0; i<numrects; ++i ) {
      short *srcBits, *dstBits;
      int w,h;
      srcBits = gfx_buffer + rects[i].x + rects[i].y * gfx_width;
      dstBits = ( bvdd_current_mem 
		  + ((bvdd_width - gfx_width) >> 1) 
		  + ((bvdd_height - gfx_height) >> 1) * bvdd_width
		  + rects[i].x + rects[i].y * bvdd_width );
      w = rects[i].w << 1;
      h = rects[i].h;

      while (h--) {
	memcpy(dstBits, srcBits, w);
	dstBits += bvdd_width;
	srcBits += gfx_width;
      }
    }
  }

  // rotation 270 degree
  static void QT_NormalUpdate3(_THIS, int numrects, SDL_Rect *rects)
  {
    updateRect_num++;
    if(updateRect_num >15 && flip_num < 5 ){    // App has no SDL_FLIP
     
      if( (_this->screen->flags & SDL_DOUBLEBUF) && !no_flip){
	st_bvdd_frame.plane_id = base_frame[0].id;
	ioctl(bvdd_fd, BVDD_IOCTL_BRANCH_BASE_FRAME, &st_bvdd_frame);
	/*
	if ( ioctl(bvdd_fd, BVDD_IOCTL_BRANCH_BASE_FRAME, &st_bvdd_frame)< 0 ) {
	  SDL_SetError("QT_UpdateRects: BVDD ioctl failed");
	  return;
	}
	*/
	bvdd_current_mem = (short*)bvdd_mapped_mem[0];
	printf("QT_UpdateRects: Not use SDL_Flip.\n");
	no_flip = 1;
      }
    }

    //printf("Rect %d flip %d\n",updateRect_num,flip_num);

    for(int i=0; i<numrects; ++i ) {

      if (rects[i].w != bvdd_width || rects[i].h != bvdd_height) {
	// partial update
	const short *pSrc = gfx_buffer + (rects[i].x + rects[i].y * gfx_width);
	int dstBorderLeftRight = (bvdd_width - gfx_height) >>1;
	int dstBorderTopBottom = (bvdd_height - gfx_width) >>1;
	int dstStartOffsetX = bvdd_width - dstBorderLeftRight - 1;
	int dstStartOffsetY = dstBorderTopBottom;
	short *pDst = bvdd_current_mem +
	  (dstStartOffsetY * bvdd_width + dstStartOffsetX);
	pDst += rects[i].x * bvdd_width;
	pDst -= rects[i].y;
	
	int width = rects[i].w, height = rects[i].h, w;
	int srcYAdd = gfx_width - rects[i].w;
	int dstXAdd = bvdd_width;
	int dstYSub = (bvdd_width * rects[i].w) + 1;
	

	while (height--) {
	  w = width;
	  while (w--) {
	    *pDst = *pSrc ++;
	    pDst += dstXAdd;
	  }
	  pSrc += srcYAdd;
	  pDst -= dstYSub;
	}

	// 		fprintf(stderr, "---- Blit end\n");
      } else {
	// full update
	const short *src = gfx_buffer;
	short *dst = bvdd_current_mem + (bvdd_width - 1);
	int w = rects[i].w;
	int h = rects[i].h;
	int i, dstSub = bvdd_width * w + 1;
	int dstAdd = bvdd_width;
	
	while (h--) {
	  i = w;
	  while (i--) {
	    *dst = *src ++;
	    dst += dstAdd;
	  }
	  dst -= dstSub;
	}
      }

    }

  }

  // flip screen for double buffer
  static int QT_FlipHWSurface(_THIS, SDL_Surface *current)
  {
    flip_num++;

    if( no_updateRects || (updateRect_num <5 && flip_num > 10) ){    // App has no SDL_UpdateRects

      SDL_UpdateRect(_this->screen, 0, 0, 0, 0);

      if(!no_updateRects){
	no_updateRects = 1;
	printf("QT_Flip: No UpdateRect and SDL_DOUBLEBUF enable app.\n");
      }

    }

    //printf("Rect %d flip %d\n",updateRect_num,flip_num);

    st_bvdd_frame.plane_id = base_frame[flip_page].id;

    ioctl(bvdd_fd, BVDD_IOCTL_BRANCH_BASE_FRAME, &st_bvdd_frame);
    /*
    if ( ioctl(bvdd_fd, BVDD_IOCTL_BRANCH_BASE_FRAME, &st_bvdd_frame)< 0 ) {
      fprintf(stderr,"ioctl failed");
      return(-1);
    }
    */
    if(no_flip)
      flip_page = 0;
    else{
      flip_page++;
      flip_page %= 3;
    }
    if(sdlRotation == -1){
      current->pixels = (short*)bvdd_mapped_mem[flip_page];
    }else{
      bvdd_current_mem = (short*)bvdd_mapped_mem[flip_page];
    }
    return(0);
  }

  static void QT_DirectUpdate(_THIS, int numrects, SDL_Rect *rects)
  {
    if(updateRect_num >10 || flip_num < 4){    // App has no SDL_FLIP

      if(!no_flip){
	st_bvdd_frame.plane_id = base_frame[0].id;
	ioctl(bvdd_fd, BVDD_IOCTL_BRANCH_BASE_FRAME, &st_bvdd_frame);
	/* 
	if ( ioctl(bvdd_fd, BVDD_IOCTL_BRANCH_BASE_FRAME, &st_bvdd_frame)< 0 ) {
	  SDL_SetError("QT_UpdateRects: BVDD ioctl failed");
	  return;
	}
	*/
	_this->screen->pixels = (short*)bvdd_mapped_mem[flip_page];
	no_flip = 1;
      }
    }

  }

  /* Is the system palette settable? */
  int QT_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
  {
    return -1;
  }

  void QT_VideoQuit(_THIS)
  {
    // This is dumb, but if I free this, the app doesn't exit correctly.
    // Of course, this will leak memory if init video is done more than once.
    // Sucks but such is life.
    
    //    -- David Hedbor
    //    delete SDL_Win; 
    //    SDL_Win = 0

    printf("exit SDL\n");
     
    if ( bvdd_fd > 0 ) {
    	/* Restore the original video mode and palette */
    	/* We're all done with the framebuffer */
    	close(bvdd_fd);
    	bvdd_fd = -1;
    }

    if(st_lcd_mode == LCD_SHARP_QVGA)
      change_resolution(LCD_SHARP_VGA);

    _this->screen->pixels = NULL;


    QT_GrabInput(_this, SDL_GRAB_OFF);


    // Qt screen redraw for taskbar
    SDL_Win->toggleSuspend(true);
    SDL_Win->setFixedSize(qApp->desktop()->size());
    SDL_Win->showNormal();
    SDL_Win->setWFlags(Qt::WStyle_Customize | Qt::WStyle_NoBorder);
    SDL_Win->showFullScreen();


    if (isKHloaded)
	system("qcop QPE/KeyHelper \"reload()\""); //QcopEnvelope can't use. 

  }

  static int QT_IconifyWindow(_THIS) {
    SDL_Win->hide();
    
    return true;
  }

  static SDL_GrabMode QT_GrabInput(_THIS, SDL_GrabMode mode) {
    if(mode == SDL_GRAB_OFF) {
      QPEApplication::grabKeyboard();
      qApp->processEvents();
      QPEApplication::ungrabKeyboard();
    } else {
      QPEApplication::grabKeyboard();
    }
    qApp->processEvents();
    return mode;
  }

}; /* Extern C */

static void change_resolution(int mode)
{
    int xres, yres;
    int fd;

    struct fb_var_screeninfo vinfo;

    switch (mode) {
    case LCD_SHARP_QVGA:
        xres = 240;
        yres = 320;
        break;
    case LCD_SHARP_VGA:
        xres = 480;
        yres = 640;
        break;
    default:
        return;
    }

    fd = open("/dev/fb0", O_RDWR);
    if (fd < 0)
        return;

    ioctl(fd, FBIOGET_VSCREENINFO, &vinfo);
    vinfo.xres = xres;
    vinfo.xres_virtual = xres;
    vinfo.yres = yres;
    vinfo.yres_virtual = yres;
    ioctl(fd, FBIOPUT_VSCREENINFO, &vinfo);

    close(fd);

    st_lcd_mode = mode;
}

