/* -*- mode: c++; tab-width: 8; c-file-style: "gnu" -*- */
/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998, 1999, 2000, 2001  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@libsdl.org
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_QWin.cc,v 1.3 2002/09/01 00:37:24 slouken Exp $";
#endif
#include <stdio.h>
#include "SDL_QWin.h"
#include <qapplication.h>
#include <qpe/qpeapplication.h>
#include <qdirectpainter_qws.h>

extern int fb_hwrot;
extern int fb_direct;
extern bool is_VGA_machine;
static int mouse_button_mode=1;

screenRotationT screenRotation = SDL_QT_NO_ROTATION;

SDL_QWin::SDL_QWin(const QSize& size)
  : QWidget(0, "SDL_main"), my_painter(0), my_image(0),
    my_inhibit_resize(false), my_mouse_pos(-1,-1), my_flags(0),
    my_has_fullscreen(false), my_locked(0),
    rotation_(0), qteRotation_(0), keyRotation_(0), qteKeyRotation_(0),
    isSLA300InputFix_(0)
{
  for (int i = 0; i < 5; i++) curAxis_[i] = 0;
  setBackgroundMode(NoBackground);
}

SDL_QWin::~SDL_QWin() {
  // Nothing to do yet.
  if(my_image) {
    delete my_image;
  }
}

void SDL_QWin::setImage(QImage *image) {
  if ( my_image ) {
    delete my_image;
  }
  my_image = image;
  //  setFixedSize(image->size());
}

bool SDL_QWin::event(QEvent *e)
{
  if( e->type() == QEvent::WindowDeactivate ){
    QKeyEvent *ke = new QKeyEvent(QEvent::KeyPress, Qt::Key_Escape,
				  0, 0);
    QueueKey(ke, 1);
    return true;
  }

  return QWidget::event(e);

}

void SDL_QWin::resizeEvent(QResizeEvent *e) {
  if(size() != qApp->desktop()->size()) {
    // Widget is not the correct size, so do the fullscreen magic
    my_has_fullscreen = false;
    enableFullscreen();
  }
  if(my_inhibit_resize) {
    my_inhibit_resize = false;
  } else {
    SDL_PrivateResize(e->size().width(), e->size().height());
  }
}

void SDL_QWin::focusInEvent(QFocusEvent *) {
  // Always do it here, no matter the size.
  enableFullscreen();
  SDL_PrivateAppActive(true, SDL_APPINPUTFOCUS);
}

void SDL_QWin::focusOutEvent(QFocusEvent *) {
  my_has_fullscreen = false;
  SDL_PrivateAppActive(false, SDL_APPINPUTFOCUS);
}

void SDL_QWin::closeEvent(QCloseEvent *e) {
  SDL_PrivateQuit();
  e->ignore();
}

void SDL_QWin::setMousePos(const QPoint &pos) {
#if 0
  if(my_image->width() == height()) {
    if (screenRotation == SDL_QT_ROTATION_90)
      my_mouse_pos = QPoint(height()-pos.y(), pos.x());
    else if (screenRotation == SDL_QT_ROTATION_270)
      my_mouse_pos = QPoint(pos.y(), width()-pos.x());
  } else {
    my_mouse_pos = pos;
  }
#else
  QPoint realPos;

  if (is_VGA_machine) {
	realPos = pos;
	my_mouse_pos = realPos;
  }else {
	  switch (qteRotation_) {
	  case 0:
	    realPos = pos;
	    break;
	  case 1:
	    realPos.setX(pos.y());
	    realPos.setY(fbSize_.height() - 1 - pos.x());
	    break;
	  case 2:
	    realPos.setX(fbSize_.width() - 1 - pos.x());
	    realPos.setY(fbSize_.height() - 1 - pos.y());
	    break;
	  case 3:
	    realPos.setX(fbSize_.width() - 1 - pos.y());
	    realPos.setY(pos.x());
	    break;
	  };
	
	//   fprintf(stderr, "setMousePos: realPos(%d, %d)\n", realPos.x(), realPos.y());

	  int borderWidth = (fbSize_.width() - my_image->width()) >> 1;
	  int borderHeight = (fbSize_.height() - my_image->height()) >> 1;
	  if ( rotation_ & 1 ){
	    borderWidth = (fbSize_.height() - my_image->width()) >> 1;
	    borderHeight = (fbSize_.width() - my_image->height()) >> 1;
	  }
	  realPos.setX(realPos.x() - borderWidth);
	  realPos.setY(realPos.y() - borderHeight);

	   fprintf(stderr, "setMousePos: realPos2(%d, %d)\n", realPos.x(), realPos.y());

	  switch (rotation_) {
	  case 0:
	    my_mouse_pos = realPos;
	    break;
	  case 1:
	    my_mouse_pos.setX(realPos.y());
	    my_mouse_pos.setY(my_image->height() - 1 - realPos.x());
	    break;
	  case 2:
	    my_mouse_pos.setX(my_image->width() - 1 - realPos.x());
	    my_mouse_pos.setY(my_image->height() - 1 - realPos.y());
	    break;
	  case 3:
	    my_mouse_pos.setX(realPos.y());//my_image->width() - 1 - realPos.y());
	    my_mouse_pos.setY(my_image->height() - realPos.x());
	    break;
	  }
  }

//   fprintf(stderr, "setMousePos: mymousePos(%d, %d)\n", (short)my_mouse_pos.x(), (short)my_mouse_pos.y());

#endif
}

void SDL_QWin::mouseMoveEvent(QMouseEvent *e) {
  Qt::ButtonState button = e->button();
  int sdlstate = 0;
  if( (button & Qt::LeftButton)) {
    sdlstate |= SDL_BUTTON_LMASK;
  }
  if( (button & Qt::RightButton)) {
    sdlstate |= SDL_BUTTON_RMASK;
  }
  if( (button & Qt::MidButton)) {
    sdlstate |= SDL_BUTTON_MMASK;
  }
  setMousePos(e->pos());
  SDL_PrivateMouseMotion(sdlstate, 0, my_mouse_pos.x(), my_mouse_pos.y());
}

void SDL_QWin::mousePressEvent(QMouseEvent *e) {
  mouseMoveEvent(e);
  Qt::ButtonState button = e->button();
  SDL_PrivateMouseButton(SDL_PRESSED,mouse_button_mode,
			 my_mouse_pos.x(), my_mouse_pos.y());
}

void SDL_QWin::mouseReleaseEvent(QMouseEvent *e) {
  setMousePos(e->pos());
  Qt::ButtonState button = e->button();
  SDL_PrivateMouseButton(SDL_RELEASED,mouse_button_mode,
			 my_mouse_pos.x(), my_mouse_pos.y());
  my_mouse_pos = QPoint(-1, -1);
}

#ifndef __i386__

static void
blitRotate3(Uint16 *dstBits, const QSize& dstSize,
            const Uint16 *srcBits, const QSize& srcSize,
            const QRect& srcBlitRect);

static void
blitRotate0(Uint16 *dstBits, const QSize& dstSize,
            const Uint16 *srcBits, const QSize& srcSize,
            const QRect& srcBlitRect)
{
  srcBits += srcBlitRect.left() + srcBlitRect.top() * srcSize.width();
  dstBits += ((dstSize.width() - srcSize.width()) >> 1)
    + (((dstSize.height() - srcSize.height()) >> 1) * dstSize.width());
  dstBits += srcBlitRect.left() + srcBlitRect.top() * dstSize.width();
  int w = srcBlitRect.width() << 1;
  int h = srcBlitRect.height();

  while (h--) {
    memcpy(dstBits, srcBits, w);
    dstBits += dstSize.width();
    srcBits += srcSize.width();
  }
}

static void
blitRotate1(Uint16 *dstBits, const QSize& dstSize,
            const Uint16 *srcBits, const QSize& srcSize,
            const QRect& srcBlitRect)
{
//   fprintf(stdout, "SDL_QT_ROTATION (1) is not implemented.\n");
  blitRotate3(dstBits, dstSize, srcBits, srcSize, srcBlitRect);
}

static void
blitRotate2(Uint16 *dstBits, const QSize& dstSize,
            const Uint16 *srcBits, const QSize& srcSize,
            QRect srcBlitRect)
{
//   if (srcBlitRect.left() & 1)
//     srcBlitRect.setLeft(srcBlitRect.left() - 1);
//   if (srcBlitRect.right() & 1)
//     srcBlitRect.setRight(srcBlitRect.right() - 1);

  int dstBorderLeftRight = (dstSize.width() - srcSize.width()) >>1;
  int dstBorderTopBottom = (dstSize.height() - srcSize.height()) >>1;
  int dstStartOffsetX = dstSize.width() - dstBorderLeftRight - 2;
  int dstStartOffsetY = dstSize.height() - dstBorderTopBottom - 1;
  int left = srcBlitRect.left() & ~(int)1; // to be 32bit aligned
  srcBits += left + srcBlitRect.top() * srcSize.width();
  dstBits += dstStartOffsetY * dstSize.width() + dstStartOffsetX;
  dstBits -= left;
  dstBits -= srcBlitRect.top() * dstSize.width();

  Uint32 *pSrc = (Uint32 *)srcBits;
  Uint32 *pDst = (Uint32 *)dstBits;

  int width = srcBlitRect.width(), width2 = srcBlitRect.width(), w;
  if ( srcBlitRect.left() & 1 ){
    width--; width2++;
  }
  if ( (srcBlitRect.left()+srcBlitRect.width()) & 1 ){
    width--; width2++;
  }
  width >>= 1;
  int height = srcBlitRect.height();
  int srcYAdd = srcSize.width() - width2;
  int dstYSub = dstSize.width() - width2;

  while (height--) {
    w = width;
    if (srcBlitRect.left() & 1){
      *((Uint16 *)pDst) = *(((Uint16 *)pSrc) + 1);
      pDst--; pSrc++;
    }
    while (w--) {
      *pDst-- = (*pSrc << 16) | (*pSrc >> 16);
      ++pSrc;
    }
    if ((srcBlitRect.left()+srcBlitRect.width()) & 1){
      *(((Uint16 *)pDst) + 1) = *((Uint16 *)pSrc);
      pDst--; pSrc++;
    }
    (Uint16 *)pSrc += srcYAdd;
    (Uint16 *)pDst -= dstYSub;
  }
}

#define BLIT_ROTATE_3_MOV32

static void
blitRotate3(Uint16 *dstBits, const QSize& dstSize,
            const Uint16 *srcBits, const QSize& srcSize,
            const QRect& srcBlitRect)
{
  if (srcBlitRect.width() != dstSize.height() ||
      srcBlitRect.height() != dstSize.width()) {
    // partial update
    const Uint16 *pSrc = srcBits +
      (srcBlitRect.left() + srcBlitRect.top() * srcSize.width());

    int dstBorderLeftRight = (dstSize.width() - srcSize.height()) >>1;
    int dstBorderTopBottom = (dstSize.height() - srcSize.width()) >>1;
    int dstStartOffsetX = dstSize.width() - dstBorderLeftRight - 1;
    int dstStartOffsetY = dstBorderTopBottom;
    Uint16 *pDst = dstBits +
      (dstStartOffsetY * dstSize.width() + dstStartOffsetX);
    pDst += srcBlitRect.left() * dstSize.width();
    pDst -= srcBlitRect.top();

    int width = srcBlitRect.width(), height = srcBlitRect.height(), w;
    int srcYAdd = srcSize.width() - srcBlitRect.width();
    int dstXAdd = dstSize.width();
    int dstYSub = (dstSize.width() * srcBlitRect.width()) + 1;

#if 0
    fprintf(stderr, "---- Blit begin\n");
    fprintf(stderr, "srcSize.width: %d srcSize.height:%d\n",
            srcSize.width(), srcSize.height());
    fprintf(stderr, "srcBlitRect.left:%d srcBlitRect.right:%d srcBlitRect.top:%d srcBlitRect.bottom:%d srcBlitRect.width:%d srcBlitRect.height:%d\n",
            srcBlitRect.left(), srcBlitRect.right(), srcBlitRect.top(), srcBlitRect.bottom(), srcBlitRect.width(), srcBlitRect.height());

    fprintf(stderr, "dstSize.width: %d dstSize.height:%d\n",
            dstSize.width(), dstSize.height());
    fprintf(stderr, "dstBorderLeftRight:%d dstBorderTopBottom:%d dstStartOffsetX:%d dstStartOffsetY:%d\n",
            dstBorderLeftRight, dstBorderTopBottom, dstStartOffsetX, dstStartOffsetY);
    fprintf(stderr, "srcYAdd:%d dstXAdd:%d dstYSub:%d\n", srcYAdd, dstXAdd, dstYSub);
#endif

    while (height--) {
      w = width;
      while (w--) {
        *pDst = *pSrc ++;
        pDst += dstXAdd;
      }
      pSrc += srcYAdd;
      pDst -= dstYSub;
    }
    // 		fprintf(stderr, "---- Blit end\n");
  } else {
    // full update
#if !defined(BLIT_ROTATE_3_MOV32)
    const Uint16 *src = srcBits;
    Uint16 *dst = dstBits + (dstSize.width() - 1);
    int w = srcBlitRect.width();
    int h = srcBlitRect.height();
    int i, dstSub = dstSize.width() * w + 1;
    int dstAdd = dstSize.width();

    while (h--) {
      i = w;
      while (i--) {
        *dst = *src ++;
        dst += dstAdd;
      }
      dst -= dstSub;
    }
#else  // BLIT_ROTATE_3_MOV32
    Uint32 *src1 = (Uint32 *)(srcBits);
    Uint32 *src2 = (Uint32 *)(srcBits + srcSize.width());
    Uint32 *dst1 = (Uint32 *)(dstBits + (dstSize.width() - 2));
    Uint32 *dst2 = (Uint32 *)(dstBits + (dstSize.width() + dstSize.width() - 2));
    int w = srcBlitRect.width() >> 3;
    int h = srcBlitRect.height() >> 1;
    int i, dstSub = ((dstSize.width() * srcBlitRect.width()) >> 1) + 1;
    int dstAdd = dstSize.width();

    Uint32 a, b;
    while (h--) {
      i = w;
      while (i--) {
        a = *src1 ++;
        b = *src2 ++;
        *dst1 = (a << 16) | (b & 0xFFFF);
        *dst2 = (a & 0xFFFF0000) | (b >> 16);
        dst1 += dstAdd;
        dst2 += dstAdd;

        a = *src1 ++;
        b = *src2 ++;
        *dst1 = (a << 16) | (b & 0xFFFF);
        *dst2 = (a & 0xFFFF0000) | (b >> 16);
        dst1 += dstAdd;
        dst2 += dstAdd;

        a = *src1 ++;
        b = *src2 ++;
        *dst1 = (a << 16) | (b & 0xFFFF);
        *dst2 = (a & 0xFFFF0000) | (b >> 16);
        dst1 += dstAdd;
        dst2 += dstAdd;

        a = *src1 ++;
        b = *src2 ++;
        *dst1 = (a << 16) | (b & 0xFFFF);
        *dst2 = (a & 0xFFFF0000) | (b >> 16);
        dst1 += dstAdd;
        dst2 += dstAdd;
      }
      src1 += srcSize.width() >> 1;
      src2 += srcSize.width() >> 1;
      dst1 -= dstSub;
      dst2 -= dstSub;
    }
#endif // BLIT_ROTATE_3_MOV32
  }
}

#endif // __i386__

void SDL_QWin::repaintRect(const QRect& rect) {
  if(!my_painter || !rect.width() || !rect.height()) {
    return;
  }
#ifndef __i386__
  if (QPixmap::defaultDepth() == 16 && my_painter->numRects() >= 0) {
    Uint16 *fb = (Uint16*)my_painter->frameBuffer();
    Uint16 *buf = (Uint16*)my_image->bits();
    switch (rotation_) {
    case 0:
      blitRotate0(fb, fbSize_,
                  buf, QSize(my_image->width(), my_image->height()),
                  rect);
      break;
    case 1:
      blitRotate1(fb, fbSize_,
                  buf, QSize(my_image->width(), my_image->height()),
                  rect);
      break;
    case 2:
      blitRotate2(fb, fbSize_,
                  buf, QSize(my_image->width(), my_image->height()),
                  rect);
      break;
    case 3:
      blitRotate3(fb, fbSize_,
                  buf, QSize(my_image->width(), my_image->height()),
                  rect);
      break;
    }
  } else {
#endif // __i386__
    QDirectPainter pp(this);
    pp.drawImage(rect.topLeft(), *my_image, rect);
//     pp.end();
#ifndef __i386__
  }
#endif // __i386__
}

// This paints the current buffer to the screen, when desired. 
void SDL_QWin::paintEvent(QPaintEvent *ev) {  
  if(my_image && isVisible() && isActiveWindow()) {
    // TODO: better handling
    lockScreen(true);
    repaintRect(ev->rect());
    unlockScreen();
  }
}  

int SDL_QWin::ApplyKeyRotation(int key)
{
  int c;
  int sdlScancode[] = { SDLK_LEFT, SDLK_DOWN, SDLK_RIGHT, SDLK_UP };

  switch (key) {
  case Qt::Key_Left:
    c = 0;
    break;
  case Qt::Key_Down:
    c = 1;
    break;
  case Qt::Key_Right:
    c = 2;
    break;
  case Qt::Key_Up:
    c = 3;
    break;
  default:
    return 0;
  }

//   c = (c + qteKeyRotation_) & 3;
//   return sdlScancode[(c - keyRotation_) & 3];
  return sdlScancode[(c + qteKeyRotation_ - keyRotation_) & 3];
}

/* Function to translate a keyboard transition and queue the key event
 * This should probably be a table although this method isn't exactly
 * slow.
 */
void SDL_QWin::QueueKey(QKeyEvent *e, int pressed)
{
  if (e->isAutoRepeat())
    return;

  SDL_keysym keysym;
  int scancode = e->key();

  /* Set the keysym information */
  if(scancode >= 'A' && scancode <= 'Z') {
    // Qt sends uppercase, SDL wants lowercase
    keysym.sym = static_cast<SDLKey>(scancode + 32);
  } else if(scancode  >= 0x1000) {
    // Special keys
    switch(scancode) {
    case Qt::Key_Escape: scancode = SDLK_ESCAPE; break;
    case Qt::Key_Tab: scancode = SDLK_TAB; break;
    case Qt::Key_Backspace: scancode = SDLK_BACKSPACE; break;
    case Qt::Key_Return: scancode = SDLK_RETURN; break;
    case Qt::Key_Enter: scancode = SDLK_KP_ENTER; break;
    case Qt::Key_Insert: scancode = SDLK_INSERT; break;
    case Qt::Key_Delete: scancode = SDLK_DELETE; break;
    case Qt::Key_Pause: scancode = SDLK_PAUSE; break;
    case Qt::Key_Print: scancode = SDLK_PRINT; break;
    case Qt::Key_SysReq: scancode = SDLK_SYSREQ; break;
    case Qt::Key_Home: scancode = SDLK_HOME; break;
    case Qt::Key_End: scancode = SDLK_END; break;
    // We want the control keys to rotate with the screen
    case Qt::Key_Left:
    case Qt::Key_Up:
    case Qt::Key_Right:
    case Qt::Key_Down:
      scancode = ApplyKeyRotation(scancode);
      break;
    case Qt::Key_Prior: scancode = SDLK_PAGEUP; break;
    case Qt::Key_Next: scancode = SDLK_PAGEDOWN; break;
    case Qt::Key_Shift: scancode = SDLK_LSHIFT; break;
    case Qt::Key_Control: scancode = SDLK_LCTRL; break;
    case Qt::Key_Meta: scancode = SDLK_LMETA; break;
    case Qt::Key_Alt: scancode = SDLK_LALT; break;
    case Qt::Key_CapsLock: scancode = SDLK_CAPSLOCK; break;
    case Qt::Key_NumLock: scancode = SDLK_NUMLOCK; break;
    case Qt::Key_ScrollLock: scancode = SDLK_SCROLLOCK; break;
    case Qt::Key_F1: scancode = SDLK_F1; break;
    case Qt::Key_F2: scancode = SDLK_F2; break;
    case Qt::Key_F3: scancode = SDLK_F3; break;
    case Qt::Key_F4: scancode = SDLK_F4; break;
    case Qt::Key_F5: scancode = SDLK_F5; break;
    case Qt::Key_F6: scancode = SDLK_F6; break;
    case Qt::Key_F7: scancode = SDLK_F7; break;
    case Qt::Key_F8: scancode = SDLK_F8; break;
    case Qt::Key_F9: scancode = SDLK_F9; break;
    case Qt::Key_F10: 
      scancode = SDLK_F10;
      mouse_button_mode=1;
      break;
    case Qt::Key_F11:
      scancode = SDLK_F11;
      mouse_button_mode=3;
      break;
    case Qt::Key_F12:
      scancode = SDLK_F12;
      mouse_button_mode=2;
      break;
    case Qt::Key_F13: scancode = SDLK_F13; break;
    case Qt::Key_F14: scancode = SDLK_F14; break;
    case Qt::Key_F15: scancode = SDLK_F15; break;
    case Qt::Key_Super_L: scancode = SDLK_LSUPER; break;
    case Qt::Key_Super_R: scancode = SDLK_RSUPER; break;
    case Qt::Key_Menu: scancode = SDLK_MENU; break;
    case Qt::Key_Help: scancode = SDLK_HELP; break;

    case 0x5001:                                //zaurus zen key Qt::Key_F21
    case 0x5002: scancode = SDLK_LSUPER; break; //zaurus zen key Qt::Key_F25
    case 0x5003:                                //zaurus kana key Qt::Key_F26
    case 0x5004: scancode = SDLK_RSUPER; break; //zaurus kana key Qt::Key_F31
    case Qt::Key_F22: scancode = SDLK_MENU; break;   //zaurus Fn key

    case Qt::Key_F33:
      // FIXME: This is a hack to enable the OK key on
      // Zaurii devices. SDLK_RETURN is a suitable key to use
      // since it often is used as such.
      //     david@hedbor.org
      scancode = SDLK_RETURN;
      break;

    default:
      scancode = SDLK_UNKNOWN;
      break;
    }
    keysym.sym = static_cast<SDLKey>(scancode);
  } else {
    keysym.sym = static_cast<SDLKey>(scancode);
  }
  keysym.scancode = scancode;
  keysym.mod = KMOD_NONE;
  ButtonState st = e->state();
  if( (st & ShiftButton) )   { keysym.mod = static_cast<SDLMod>(keysym.mod | KMOD_LSHIFT);  }
  if( (st & ControlButton) ) { keysym.mod = static_cast<SDLMod>(keysym.mod | KMOD_LCTRL);  }
  if( (st & AltButton) )     { keysym.mod = static_cast<SDLMod>(keysym.mod | KMOD_LALT);  }
  if ( SDL_TranslateUNICODE ) {
    QChar qchar = e->text()[0];
    keysym.unicode = qchar.unicode();
  } else {
    keysym.unicode = 0;
  }

  /* NUMLOCK and CAPSLOCK are implemented as double-presses in reality */
  //	if ( (keysym.sym == SDLK_NUMLOCK) || (keysym.sym == SDLK_CAPSLOCK) ) {
  //		pressed = 1;
  //	}

  if (isSLA300InputFix_ &&
      (keysym.sym == SDLK_SPACE || (keysym.sym >= 273 && keysym.sym <= 276))) {
      if (keysym.sym >= 273) curAxis_[keysym.sym-273] = pressed;
      else curAxis_[4] = pressed;
  }

  /* Queue the key event */
  if ( pressed ) {
      /* fprintf(stderr, "press %d\n", keysym.sym); */
      SDL_PrivateKeyboard(SDL_PRESSED, &keysym);
      if (isSLA300InputFix_ &&
          (keysym.sym >= 273 && keysym.sym <= 276)) {
          if (keysym.sym == SDLK_UP) {
              if (curAxis_[1]) {
                  keysym.sym = SDLK_DOWN;
                  keysym.scancode = Qt::Key_Down;
                  curAxis_[1] = 0;
                  /* fprintf(stderr, "force release %d\n", keysym.sym); */
                  SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
              }
          }
          else if (keysym.sym == SDLK_DOWN) {
              if (curAxis_[0]) {
                  keysym.sym = SDLK_UP;
                  keysym.scancode = Qt::Key_Up;
                  curAxis_[0] = 0;
                  /* fprintf(stderr, "force release %d\n", keysym.sym); */
                  SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
              }
          }
          else if (keysym.sym == SDLK_RIGHT) {
              if (curAxis_[3]) {
                  keysym.sym = SDLK_LEFT;
                  keysym.scancode = Qt::Key_Left;
                  curAxis_[3] = 0;
                  /* fprintf(stderr, "force release %d\n", keysym.sym); */
                  SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
              }
          }
          else if (keysym.sym == SDLK_LEFT) {
              if (curAxis_[2]) {
                  keysym.sym = SDLK_RIGHT;
                  keysym.scancode = Qt::Key_Right;
                  curAxis_[2] = 0;
                  /* fprintf(stderr, "force release %d\n", keysym.sym); */
                  SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
              }
          }
      }
  } else {
      /* fprintf(stderr, "release %d\n", keysym.sym); */
      SDL_PrivateKeyboard(SDL_RELEASED, &keysym);

      if (isSLA300InputFix_ &&
          (keysym.sym == SDLK_SPACE ||
           (keysym.sym >= 273 && keysym.sym <= 276))) {
          for (int i = 0; i < 4; i++) {
              if (curAxis_[i]) {
                  int sym = i+273;
                  keysym.sym = static_cast<SDLKey>(sym);
                  if (sym == SDLK_UP) keysym.scancode = Qt::Key_Up;
                  else if (sym == SDLK_RIGHT) keysym.scancode = Qt::Key_Right;
                  else if (sym == SDLK_DOWN) keysym.scancode = Qt::Key_Down;
                  else if (sym == SDLK_LEFT) keysym.scancode = Qt::Key_Left;
                  curAxis_[i] = 0;
                  /* fprintf(stderr, "force release %d\n", keysym.sym); */
                  SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
              }
          }
          if (curAxis_[4]) {
              keysym.scancode = keysym.sym = SDLK_SPACE;
              curAxis_[4] = 0;
              /* fprintf(stderr, "force release %d\n", keysym.sym); */
              SDL_PrivateKeyboard(SDL_RELEASED, &keysym);
          }
      }
  }
}

void SDL_QWin::setFullscreen(bool fs_on) {
  my_has_fullscreen = false;
  enableFullscreen();
}

void SDL_QWin::enableFullscreen() {
  // Make sure size is correct
  if(!my_has_fullscreen) {
    setFixedSize(qApp->desktop()->size());
    // This call is needed because showFullScreen won't work
    // correctly if the widget already considers itself to be fullscreen.
    showNormal();
    // This is needed because showNormal() forcefully changes the window
    // style to WSTyle_TopLevel.
    setWFlags(WStyle_Customize | WStyle_NoBorder);
    // Enable fullscreen.
    showFullScreen();
    my_has_fullscreen = true;
  }
}

bool SDL_QWin::lockScreen(bool force) {
  if(!my_painter) {
    if(force || (isVisible() && isActiveWindow())) {
      my_painter = new QDirectPainter(this);
    } else {
      return false;
    }
  }
  my_locked++; // Increate lock refcount
  return true;
}

void SDL_QWin::unlockScreen() {
  if(my_locked > 0) {
    my_locked--; // decrease lock refcount;
  }
  if(!my_locked && my_painter) {
//     my_painter->end();
    delete my_painter;
    my_painter = 0;
  }
}
